package com.teletip.wado.controller;

import java.awt.image.BufferedImage;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.List;

import javax.servlet.http.HttpServletResponse;

import org.dcm4che.dict.UIDs;
import org.dcm4cheri.util.StringUtils;

import com.teletip.wado.cache.WADOCacheImpl;
import com.teletip.wado.info.WADOInfo;
import com.teletip.wado.response.WADOImageResponseObjectImpl;
import com.teletip.wado.response.WADOResponseObject;
import com.teletip.wado.response.WADOStreamResponseObjectImpl;
import com.teletip.wado.util.DcmUtil;

/**
 * WADOCtrl WADO sorgularinin yonetilip Lucene Indeks'ten alinan url bilgisi ile ilgili 
 * goruntu isleme islemlerinin WADOSupport destegi ile gerceklestirildigi kontrol sinifidir.
 * @author	Ayhan Ozan Yilmaz
 * @since	01.01.2017
 */
public class WADOCtrl {

	public WADOCtrl(String dbConn) {
	}

	private static final String ERROR_INVALID_REGION_FORMAT =
            "Error: region parameter is invalid! Must be a comma separated list of 4 decimal strings.";

    private static final String ERROR_INVALID_REGION_OUT_OF_RANGE =
            "Error: region parameter is invalid! Coordinates must be in range [0..1].";

    private static final String ERROR_INVALID_REGION_DIMENSION =
            "Error: region parameter is invalid! Width and height of specified region must be > 0.";

    private static final String ERROR_NULL_WINDOW_WIDTH =
            "Error: windowWidth parameter is invalid! Must specify a value.";

    private static final String ERROR_NULL_WINDOW_CENTER =
            "Error: windowCenter parameter is invalid! Must specify a value.";

    private static final String ERROR_INVALID_WINDOW_WIDTH_TYPE =
            "Error: windowWidth parameter is invalid! Must be a decimal string.";

    private static final String ERROR_INVALID_WINDOW_CENTER_TYPE =
            "Error: windowCenter parameter is invalid! Must be a decimal string.";

    private static final String ERROR_INVALID_WINDOW_WIDTH_VALUE =
            "Error: windowWidth parameter is invalid! Width must be > 0.";

    private static final String ERROR_INVALID_IMAGE_QUALITY_TYPE =
            "Error: imageQuality parameter is invalid! Must be a integer string.";

    private static final String ERROR_INVALID_IMAGE_QUALITY_VALUE =
            "Error: imageQuality parameter is invalid! Quality must be in range [1..100].";
    
    public static final int OK = 0;

    public static final int INVALID_WADO_URL = -1;
    public static final int INVALID_ROWS = -2;
    public static final int INVALID_COLUMNS = -3;
    public static final int INVALID_FRAME_NUMBER = -4;
    public static final int INVALID_CONTENT_TYPE = -5;
    public static final int INVALID_TRANSFER_SYNTAX = -6;
    public static final int INVALID_REGION = -7;
    public static final int INVALID_WINDOW_LEVEL = -8;
    public static final int INVALID_IMAGE_QUALITY = -9;
    public static final int INVALID_UID = -10;
    
	public static final String OPERATION_NONE = "NONE";

	public static final String OPERATION_GET_STUDY = "getStudy";

	public static final String OPERATION_TEST_STUDY = "testStudy";

	public static final String OPERATION_GET_DICOM = "getDicom";

	public static final String OPERATION_GET_DICOM_INSTANCE_ITEMS = "getDicomInstanceItems";

	public static final String OPERATION_DOWNLOAD_STUDY = "downloadStudy";

	// request parameter names
	public static final String PARAM_STUDY_ID = "studyId";

	public static final String PARAM_STUDY_XML = "studyXml";

	public static final String PARAM_OPERATION = "op";  

	public static final String PARAM_FILE_ID = "studyFileId";

	public static final String PARAM_FILE_NAME = "studyFileName";

	public static final String PARAM_DICOM_INSTANCE_ITEM_XML = "dicomInstanceItemXml";

	public static final String PARAM_DICOM_INSTANCE_ID = "dicomInstanceId";
	
	//WADO parameters
	public static final String PARAM_REQUEST_TYPE = "requestType";
	
	public static final String PARAM_STUDY_UID = "studyUID"; 
	
	public static final String PARAM_SERIES_UID = "seriesUID"; 
	
	public static final String PARAM_OBJECT_UID = "objectUID"; 
	
	public static final String PARAM_CONTENT_TYPE = "contentType";  
	
	public static final String PARAM_CHARSET = "charset"; 
	
	public static final String PARAM_ANONYMIZE = "anonymize"; 
	
	public static final String PARAM_ANNOTATION = "annotation"; 
	
	public static final String PARAM_ROWS = "rows"; 
	
	public static final String PARAM_COLUMNS = "columns"; 
	
	public static final String PARAM_REGION = "region"; 
	
	public static final String PARAM_WINDOW_CENTER = "windowCenter";
	
	public static final String PARAM_WINDOW_WIDTH = "windowWidth"; 
	
	public static final String PARAM_FRAME_NUMBER = "frameNumber";
	
	public static final String PARAM_IMAGE_QUALITY = "imageQuality"; 
	
	public static final String PARAM_PRESENTATION_UID =  "presentationUID"; 
	
	public static final String PARAM_PRESENTATION_SERIES_UID = "presentationSeriesUID"; 
	
	public static final String PARAM_TRANSFER_SYNTAX = "transferSyntax";
	
	public static final String PARAM_HEADER_AUTHORIZATION = "Authorization";
	
	public static final String PARAM_REDIRECT = "redir";
	
	public static final String PARAM_USE_ORIG = "useOrig";
	
	public static final String PARAM_PRIVATE_TAGS = "privateTags";
	
	public static final String PARAM_SIMPLE_FRAME_LIST = "simpleFrameList";
	
	public static final String PARAM_CALCULATED_FRAME_LIST = "calculatedFrameList";

	// variable that holds page data
	WADOInfo wadoInfo = new WADOInfo();

	ByteArrayOutputStream baos;
	
	WADOResponseObject responseObject;
	
	private WADOSupport wadoSupport = new WADOSupport();

	public ByteArrayOutputStream getBaos() {
		return baos;
	}

	String resultXml = "";
	String errorMessage = "";

	public void processCtrl() {
		// local reference to dbUtil

		int iErr = checkRequest();
		if (iErr < 0) {
			return;
		}

		// --------- FIRST PART : CACHE MANAGEMENT ---------

		// --------- SECOND PART : PERFORM OPERATIONS ---------

		if (wadoInfo.getOperation().equalsIgnoreCase(OPERATION_NONE)) {
			try {
				responseObject = wadoSupport.getWADOObject(wadoInfo);
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
	}

	public WADOInfo getWadoInfo() {
		return wadoInfo;
	}

	public String getResultXml() {
		return resultXml;
	}
	
	/**
     * Checks this request object and returns an error code.
     * <p>
     * <DL>
     * <DT>Following checks:</DT>
     * <DD> requestType must be "WADO"</DD>
     * <DD> studyUID, seriesUID and objectUID must be set</DD>
     * <DD> if rows is set: check if it is parseable to int</DD>
     * <DD> if columns is set: check if it is parseable to int</DD>
     * <DD> if frameNumber is set: check if it is parseable to int</DD>
     * </DL>
     * 
     * @return OK if it is a valid WADO request or an error code.
     */
    public int checkRequest() {
        if (wadoInfo.getRequestType() == null
                || !wadoInfo.getRequestType().equalsIgnoreCase("WADO")
                || wadoInfo.getStudyUID() == null
                || wadoInfo.getSeriesUID() == null
                || wadoInfo.getObjectUID() == null || wadoInfo.getObjectUID().trim().length() == 0
        ) {
            setErrorMsg("Not a WADO URL!");
            return INVALID_WADO_URL;
        }
        if (!checkUIDs()) {
            return INVALID_UID;
        }
        if (wadoInfo.getRows() != null) {
            try {
                Integer.parseInt(wadoInfo.getRows());
            } catch (Exception x) {
                setErrorMsg("Error: rows parameter is invalid! Must be an integer string.");
                return INVALID_ROWS;
            }
        }
        if (wadoInfo.getColumns() != null) {
            try {
                Integer.parseInt(wadoInfo.getColumns());
            } catch (Exception x) {
                setErrorMsg("Error: columns parameter is invalid! Must be an integer string.");
                return INVALID_COLUMNS;
            }
        }
        if (wadoInfo.getFrameNumber() != null) {
            try {
                Integer.parseInt(wadoInfo.getFrameNumber());
            } catch (Exception x) {
                setErrorMsg("Error: frameNumber parameter is invalid! Must be an integer string.");
                return INVALID_FRAME_NUMBER;
            }
        }
        if (wadoInfo.getRegion() != null) {
            try {
                checkRegion(wadoInfo.getRegion());
            } catch (IllegalArgumentException e) {
                setErrorMsg(e.getMessage());
                return INVALID_REGION;
            }
        }

        if (wadoInfo.getWindowWidth() != null || wadoInfo.getWindowCenter() != null) {
            try {
                checkWindowLevel(wadoInfo.getWindowWidth(), wadoInfo.getWindowCenter());
            } catch (Exception x) {
                setErrorMsg(x.getMessage());
                return INVALID_WINDOW_LEVEL;
            }
        }

        if (wadoInfo.getImageQuality() != null) {
            try {
                checkImageQuality(wadoInfo.getImageQuality());
            } catch (IllegalArgumentException e) {
                setErrorMsg(e.getMessage());
                return INVALID_IMAGE_QUALITY;
            }
        }
        
        if (wadoInfo.getContentTypes() != null) {
            for (String ct : wadoInfo.getContentTypes()) {
                if (!WADOSupport.CONTENT_TYPES.contains(ct)) {
                    setErrorMsg("Invalid contentType "+ct+"! Must be one of " + WADOSupport.CONTENT_TYPES);
                    return INVALID_CONTENT_TYPE;
                }
            }
        }
        setErrorMsg(null);
        return OK;
    }

/**
     * Checks that the region string's value is valid. Throws
     * <code>IllegalArgumentException</code> if it isn't.
     * 
     * @param region
     *                String representing a rectangular region of an image
     * 
     * @return void
     * 
     */
    private void checkRegion(String region) {
        String[] ss = StringUtils.split(region, ',');
        if (ss.length != 4) {
            throw new IllegalArgumentException(ERROR_INVALID_REGION_FORMAT);
        }
        double[] ds = new double[4];
        for (int i = 0; i < ds.length; i++) {
            try {
                ds[i] = Double.parseDouble(ss[i]);
            } catch (NumberFormatException e) {
                throw new IllegalArgumentException(ERROR_INVALID_REGION_FORMAT);
            }
            if (ds[i] < 0. || ds[i] > 1.) {
                throw new IllegalArgumentException(
                        ERROR_INVALID_REGION_OUT_OF_RANGE);
            }
        }
        if (!(ds[0] < ds[2] && ds[1] < ds[3])) {
            throw new IllegalArgumentException(ERROR_INVALID_REGION_DIMENSION);
        }
    }

    /**
     * Checks that the windowWidth & windowCenter values are valid. Throws
     * <code>IllegalArgumentException</code> if either isn't.
     * 
     * @param windowWidth
     *                The value of the windowWidth WADO parameter
     * @param windowCenter
     *                The value of the windowCenter WADO parameter
     * 
     * @return void
     * 
     */
    private void checkWindowLevel(String windowWidth, String windowCenter)
            throws IllegalArgumentException {
        if (windowWidth == null)
            throw new IllegalArgumentException(ERROR_NULL_WINDOW_WIDTH);

        if (windowCenter == null)
            throw new IllegalArgumentException(ERROR_NULL_WINDOW_CENTER);

        double width = -1;

        try {
            width = Double.parseDouble(windowWidth);
        } catch (NumberFormatException e) {
            throw new IllegalArgumentException(ERROR_INVALID_WINDOW_WIDTH_TYPE);
        }

        if (width <= 0)
            throw new IllegalArgumentException(ERROR_INVALID_WINDOW_WIDTH_VALUE);

        try {
            Double.parseDouble(windowCenter);
        } catch (NumberFormatException e) {
            throw new IllegalArgumentException(ERROR_INVALID_WINDOW_CENTER_TYPE);
        }
    }

    private void checkImageQuality(String imageQuality) {

        int quality = -1;

        try {
            quality = Integer.parseInt(imageQuality);
        } catch (NumberFormatException e) {
            throw new IllegalArgumentException(ERROR_INVALID_IMAGE_QUALITY_TYPE);
        }

        if (quality <= 0 || quality > 100)
            throw new IllegalArgumentException(
                    ERROR_INVALID_IMAGE_QUALITY_VALUE);
    }
    
    private boolean checkUIDs() {
        // studyUID and seriesUID are not used in query! So check UID only if length > 2 to allow short URLs
        if (wadoInfo.getStudyUID().length() > 2 && !isUID(wadoInfo.getStudyUID())) { 
            this.setErrorMsg("Invalid studyUID parameter!");
            return false;
        }
        if (wadoInfo.getSeriesUID().length() > 2 && !isUID(wadoInfo.getSeriesUID())) {
            this.setErrorMsg("Invalid seriesUID parameter!");
            return false;
        }
        if (!isUID(wadoInfo.getObjectUID())) {
            this.setErrorMsg("Invalid objetcUID parameter!");
            return false;
        }
        if (wadoInfo.getTransferSyntax() != null && !UIDs.isValid(wadoInfo.getTransferSyntax())) {
            this.setErrorMsg("Invalid transferSyntax parameter! Not a valid UID");
            return false;
        }
        return true;
    }

    public static boolean isUID(String uid) {
        if (uid == null || uid.length() < 3 || uid.charAt(0) == '.')
            return false;
        char[] a = uid.toCharArray();
        boolean expectDigit = true;
        for (int i = 0; i < a.length; ++i) {
            if (a[i] == '.') {
                expectDigit = true;
            } else if (a[i] > '9' || a[i] < '0') {
                return false;
            } else {
                expectDigit = false;
            }
        }
        return !expectDigit;
    }

	public String getErrorMsg() {
		return errorMessage;
	}

	public void setErrorMsg(String errorMessage) {
		this.errorMessage = errorMessage;
	}

	public WADOResponseObject getResponseObject() {
		return responseObject;
	}
	
	private WADOResponseObject handleImage(File dicomFile, WADOInfo wadoInfo) {
        String rows = wadoInfo.getRows();
        String columns = wadoInfo.getColumns();
        String frameNumber = wadoInfo.getFrameNumber();
        String region = wadoInfo.getRegion();
        String windowWidth = wadoInfo.getWindowWidth();
        String windowCenter = wadoInfo.getWindowCenter();
        String imageQuality = wadoInfo.getImageQuality();
        String contentType = wadoInfo.getContentType();
        try {
            int frame = 0;
            if (frameNumber != null) {
                frame = Integer.parseInt(frameNumber)-1;
            }
                BufferedImage bi = DcmUtil.getImage(dicomFile, frame, rows, columns, region,
                        windowWidth, windowCenter, contentType, false, false, false);
                return new WADOImageResponseObjectImpl(bi, WADOCacheImpl.getWADOCache(), 
                        imageQuality != null ? imageQuality : WADOCacheImpl.getWADOCache().getImageQuality(),
                                contentType, HttpServletResponse.SC_OK, "Info: Caching disabled!");
        } catch (Exception x) {
            return new WADOStreamResponseObjectImpl(null, contentType,
                    HttpServletResponse.SC_INTERNAL_SERVER_ERROR,
                    "Unexpected error! Cant get jpeg");
        }
    }
}
