package com.teletip.wado.info;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;
import java.util.StringTokenizer;

import org.dcm4che3.data.Attributes;

import com.teletip.wado.controller.WADOCtrl;
import com.teletip.wado.controller.WADOSupport;

/**
 * WADOInfo WADO istemleri ile iletilen goruntuleme parametrelerinin kaydedildii ve 
 * WADOCtrl sinifina iletildigi uygulama katmani sinifidir. 
 * yardimci arac sinifidir.
 * @author	Ayhan Ozan Yilmaz
 * @since	01.01.2017
 */
public class WADOInfo implements Serializable {

	private static final long serialVersionUID = -1419186087573310070L;

    String operation;
    
    String studyId;
    
    String patientId;
    
    String studyXml;
    
    String patientXml;
    
    String dicomInstanceItemXml;
    
    String dicomInstanceId;
    
    //WADO parameters
    /**
     * Type of request performed. This parameter is REQUIRED. 
     * The parameter name shall be requestType. The value shall be "WADO". 
     * Note: This parameter allows other types of requests to be introduced in the future, using a similar syntax. 
     */
    String requestType;
    
    /**
     * Unique identifier of the study
     * Study Instance UID as defined in PS 3.3. This parameter is REQUIRED. 
     * The parameter name shall be studyUID. 
     * The value shall be encoded as a Unique Identifier (UID) string, as specified in PS 3.5, 
     * except that it shall not be padded to an even length with a NULL character.
     */
    String studyUID; 
    
    
    /**
     * Unique identifier of the series
     * Series Instance UID as defined in the PS 3.3. 
     * This parameter is REQUIRED. The parameter name shall be seriesUID. 
     * The value shall be encoded as a Unique Identifier (UID) string, 
     * as specified in PS 3.5, except that it shall not be padded to an even length with a NULL character. 
     */
    String seriesUID; 
    
    /**
     * Unique identifier of the object 
     * SOP Instance UID as defined in the PS 3.3. 
     * This parameter is REQUIRED.  
     * The parameter name shall be objectUID. The value shall be encoded as a unique identifier (UID) string, 
     * as specified in PS 3.5, except that it shall not be padded to an even length with a NULL character. 
     */
    String objectUID; 
    
    /**
     * MIME type(s) desired by the Web Client for the response from the Server, as defined in the IETF RFC2616. 
     * This parameter is OPTIONAL.  
     * The parameter name shall be contentType. 
     * The value shall be a list of MIME types, separated by a "," character, 
     * and potentially associated with relative degree of preference, as specified in IETF RFC2616. 
     * The Web Client shall provide list of content types it supports in the "Accept" field of the GET method. 
     * The value of the contentType parameter of the request shall be one of the values specified in that field.
     * examples: application/dicom, image/jpeg, image/gif, image/png, image/jp2  
     * default:  image/jpeg  
     */
    String contentType;  

    /**
     * Character set with which the returned object is to be encoded, as defined in the IETF RFC2616. 
     * This parameter is OPTIONAL.  
     * The parameter name shall be charset. The value shall be a list of character sets, separated by a "," character, 
     * and potentially associated with relative degree of preference, as specified in IETF RFC2616. 
     * The Web Client may provide a list of character sets it supports in the "Accept-charset" field of the GET method. 
     * If this field is present, the value of the charset parameter of the request shall be one of the values specified in it. 
     * The Server may or may not support character set conversion.
     */
    String charset; 
    
    /**
     * Removal of all patient identification information from within the DICOM object, if not already done, as defined in PS 3.15. 
     * This parameter is OPTIONAL. 
     * It shall only be present if contentType is application/dicom. 
     * This parameter is Optional The parameter name shall be anonymize. The value shall be yes. 
     * The Server may return an error if it either cannot or refuses to anonymize that object.  
     * The Server shall return a new SOP Instance UID if the content of the object has not already been anonymized.
     */
    String anonymize; 
    
    /**
     * Annotation of an object retrieved and displayed as an image. 
     * This parameter is OPTIONAL. 
     * It shall not be present if contentType is application/dicom, or is a non-image MIME type (e.g., text/*). 
     * When it is not present for an image object, no annotation may be burnt in. 
     * When used in conjunction with a presentation state object, it shall be applied after the presentation on the image. 
     * When used in conjunction with the region parameter, it shall be applied after the selection of the region. 
     * The parameter name shall be annotation. 
     * Its value is a non-empty list of one or more of the following items , separated by a "," character: 
     * "patient", for displaying patient information on the image (e.g. patient name, birth date,) 
     * "technique", for displaying technique information of the image (e.g. image number, study date, image position,). 
     * Note: The exact nature and presentation of the annotation is determined by the Server. 
     * The annotation is burned into the returned image pixels. 
     */
    String annotation; 
    
    /**
     * The parameter name shall be rows. 
     * The value shall be expressed as an integer, representing the image height to be returned.  
     * It is OPTIONAL. 
     * It shall not be present if contentType is application/dicom. 
     * If both rows and columns are specified, then each shall be interpreted as a maximum, 
     * and a size will be chosen for the image within these constraints, maintaining the correct aspect ratio. 
     * If the number of rows is absent and the number of columns is present, the number of rows shall be chosen 
     * in order to maintain the correct aspect ratio. 
     * If both are absent, the image (or selected region) is sent in its original size (or the size of the presentation state applied on the image), 
     * resulting as one pixel of screen image for each value in the image data matrix.  
     * The value shall be encoded as an integer string (IS), as specified in PS 3.5. 
     */
    String rows; 
    
    /**
     * The parameter name shall be columns. 
     * The value shall be expressed as an integer, representing the image width to be returned.  
     * It is OPTIONAL. 
     * It shall not be present if contentType is application/dicom. 
     * If both rows and columns are specified, then each shall be interpreted as a maximum, 
     * and a size will be chosen for the image within these constraints, maintaining the correct aspect ratio. 
     * If the number of columns is absent and the number of rows is present, 
     * the number of columns shall be chosen in order to maintain the correct aspect ratio. 
     * If both are absent, the image (or selected region) is sent in its original size (or the size of the presentation state applied on the image), 
     * resulting as one pixel of screen for one pixel of the image. 
     * The value shall be encoded as an integer string (IS), as specified in PS 3.5. 
     */
    String columns; 
    
    /**
     * This parameter allows selection of a rectangular region of an image matrix to be retrieved. 
     * The purpose of this parameter is to allow a user to view a selected area of the image matrix, for example at higher magnification. 
     * The parameter is OPTIONAL. 
     * The parameter name shall be region. 
     * It shall not be present if contentType is application/dicom. 
     * The value shall be expressed as a list of four positive decimal strings, separated by the ',' character, 
     * representing the region of the source image to be returned. 
     * These decimal values shall be values in a normalized coordinate system 
     * relative to the size of the original image matrix measured in rows and columns, 
     * with values ranging from 0.0 to 1.0, and representing in the following order: 
     * the x position of the top left hand corner of the region to be retrieved, 0.0 corresponding to the  first column of the image matrix. 
     * the y position of the top left hand corner of the region to be retrieved, 0.0 corresponding to the  top row of the image matrix. 
     * the x position of the bottom right hand extent of the region, 1.0 corresponding to  the last column of the image matrix, 0.0 being forbidden. 
     * the y position of the bottom right hand extent of the region, 1.0 corresponding to the last row of the image matrix, 0.0 being forbidden. 
     * Note: The Server may or may not support this parameter.
     * If this parameter is supported, an image matrix corresponding to the specified region 
     * shall be returned with size corresponding to the specified normalized coordinate values 
     * otherwise the complete image matrix shall be returned  If the presentationUID parameter is present, 
     * the region shall be selected after the corresponding presentation state has been applied on the image.
     */
    String region; 
    
    /**
     * The parameter name shall be windowCenter. Controls the luminosity of the image as defined in PS 3.3. 
     * This parameter is REQUIRED if "windowWidth" is present. 
     * This parameter shall not be present if there is a presentationUID parameter. 
     * It shall not be present if contentType is application/dicom. 
     * The value shall be encoded as a decimal string (DS), as specified in PS 3.5. 
     */
    String windowCenter;
    
    /**
     * The parameter name shall be windowWidth. 
     * Controls the contrast of the image as defined in PS 3.3. 
     * It is REQUIRED if "windowCenter" is present. 
     * This parameter shall not be present if there is a presentationUID parameter. 
     * It shall not be present if contentType is application/dicom. 
     * The value shall be encoded as a decimal string (DS), as specified in PS 3.5. 
     */
    String windowWidth; 
    
    /**
     * The parameter name shall be frameNumber". 
     * Specifies that the single frame with that number within a multi-frame image object, 
     * as defined in PS 3.3 that shall be returned. 
     * It is OPTIONAL and shall be ignored in the case of all objects other than multi-frame objects. 
     * It shall not be present if contentType is application/dicom. 
     * The value shall be encoded as an integer string (IS), as specified in PS 3.5.
     */
    String frameNumber;
    
    /**
     * The parameter name shall be imageQuality". 
     * It is OPTIONAL. 
     * It shall not be present if contentType is application/dicom, 
     * except if the transferSyntax parameter is present and corresponds to a lossy compression. 
     * If the requested MIME type is for  a lossy compressed image (e.g. image/jpeg), 
     * this parameter indicates the required quality of the image to be returned within the range 1 to 100, 100 being the best quality. 
     * Note: Decompression and recompression may degrade the image quality if the original image was already irreversibly compressed. 
     * In case the image has been already lossy compressed using the same format as required (e.g. jpeg), 
     * it may be sent as it is without decompressing and recompressing it.  
     * The value shall be encoded as an integer string (IS), as specified in PS 3.5. 
     * Note: The specific interpretation of the meaning of this parameter is left to the interpretation of the implementers of the standard.  
     */
    String imageQuality; 
    
    /**
     * The parameter name shall be presentationUID". 
     * SOP Instance UID of the presentation state storage object to be applied to the image. 
     * This parameter is OPTIONAL. 
     * It shall not be present if contentType is application/dicom. 
     * The value shall be encoded as a unique identifier (UID) string, as specified in PS 3.5, 
     * except that it shall not be padded to an even length with a NULL character. 
     * If this parameter is combined with region and/or annotation parameter(s), 
     * the presentation state shall be applied to the image prior to selecting a region and burning in annotations. 
     * If the Presentation Size Mode in the presentation state is SCALE TO FIT or TRUE SIZE, 
     * then the displayed area specified in the presentation shall be scaled to fit the size specified by the rows and columns parameters if present, 
     * otherwise the displayed area selected in the presentation state will be returned without scaling. 
     * Notes: 1. The intent of the TRUE SIZE mode in the presentation state cannot be satisfied, 
     * since the physical size of the pixels displayed by the web browser is unlikely to be known. 
     * If the Presentation Size Mode in the presentation state is MAGNIFY, 
     * then the displayed area specified in the presentation shall be magnified (scaled) as specified in the presentation state. 
     * It will then be cropped to fit the size specified by the rows and columns parameters, if present.  
     * 2. Any Displayed Area relative annotations specified in the presentation state are 
     * rendered relative to the Specified Displayed Area within the presentation state, not the size of the returned image. 
     * Though the output of the presentation state is defined in DICOM to be in P-Values 
     * (grayscale values intended for display on a device calibrated to the DICOM Grayscale Standard Display Function PS 3.14), 
     * the grayscale or color space for the image returned by the request is not defined by this standard. 
     */
    String presentationUID; 
    
    /**
     * The parameter name shall be presentationSeriesUID". 
     * Series Instance UID of the series containing the presentation state storage object to be applied on the image. 
     * This parameter is REQUIRED and shall only be present if "presentationUID" is present. 
     * The value shall be encoded as a unique identifier (UID) string, as specified in PS 3.5, 
     * except that it shall not be padded to an even length with a NULL character. 
     * Note: As specified in DICOM, the Presentation State will be in the same study as the image it applies to.  
     */
    String presentationSeriesUID; 
    
    /**
     * The parameter name shall be transferSyntax". 
     * The Transfer Syntax to be used within the DICOM image object, as specified in PS 3.6. 
     * This parameter is OPTIONAL. 
     * It shall not be present if contentType is other than application/dicom. 
     * By default the DICOM object(s) returned shall be encoded in Explicit VR Little Endian. 
     * Neither Implicit VR, nor Big Endian shall be used. 
     * The response shall be the Transfer Syntax requested if possible. 
     * If it is not possible for the response to be sent using the requested transfer syntax 
     * then the Explicit VR Little Endian Uncompressed Transfer Syntax shall be used. 
     * The value shall be encoded as an unique identifier (UID) string, as specified in PS 3.5, 
     * except that it shall not be padded to an even length with a NULL character.
     */
    String transferSyntax;
    
    private List<String> contentTypes;
    
    private Attributes objectInfo;
    
    private String privateTags;
    
    private String simpleFrameList;
    
    private String calculatedFrameList;
    
    private String redirectParam;
    
    private String authorization;
    
    private String useOrig;
    
    private String requestUrl;
    
    public String getOperation() {
        return operation;
    }

    public void setOperation(String operation) {
        this.operation = (operation != null) ? operation
                : WADOCtrl.OPERATION_NONE;
    }

	public String getStudyId() {
		return studyId;
	}

	public void setStudyId(String studyId) {
		this.studyId = studyId;
	}

	public String getPatientId() {
		return patientId;
	}

	public void setPatientId(String patientId) {
		this.patientId = patientId;
	}

	public String getStudyXml() {
		return studyXml;
	}

	public void setStudyXml(String studyXml) {
		this.studyXml = studyXml;
	}

	public String getPatientXml() {
		return patientXml;
	}

	public void setPatientXml(String patientXml) {
		this.patientXml = patientXml;
	}

	public String getRequestType() {
		return requestType;
	}

	public void setRequestType(String requestType) {
		this.requestType = requestType;
	}

	public String getDicomInstanceItemXml() {
		return dicomInstanceItemXml;
	}

	public void setDicomInstanceItemXml(String dicomInstanceItemXml) {
		this.dicomInstanceItemXml = dicomInstanceItemXml;
	}

	public String getDicomInstanceId() {
		return dicomInstanceId;
	}

	public void setDicomInstanceId(String dicomInstanceId) {
		this.dicomInstanceId = dicomInstanceId;
	}

	public String getStudyUID() {
		return studyUID;
	}

	public void setStudyUID(String studyUID) {
		this.studyUID = studyUID;
	}

	public String getSeriesUID() {
		return seriesUID;
	}

	public void setSeriesUID(String seriesUID) {
		this.seriesUID = seriesUID;
	}

	public String getObjectUID() {
		return objectUID;
	}

	public void setObjectUID(String objectUID) {
		this.objectUID = objectUID;
	}

	public String getContentType() {
		return contentType;
	}

	public void setContentType(String contentType) {
		this.contentType = contentType != null ? contentType : WADOSupport.CONTENT_TYPE_JPEG;
	}

	public String getCharset() {
		return charset;
	}

	public void setCharset(String charset) {
		this.charset = charset;
	}

	public String getAnonymize() {
		return anonymize;
	}

	public void setAnonymize(String anonymize) {
		this.anonymize = anonymize;
	}

	public String getAnnotation() {
		return annotation;
	}

	public void setAnnotation(String annotation) {
		this.annotation = annotation;
	}

	public String getRows() {
		return rows;
	}

	public void setRows(String rows) {
		this.rows = rows;
	}

	public String getColumns() {
		return columns;
	}

	public void setColumns(String columns) {
		this.columns = columns;
	}

	public String getRegion() {
		return region;
	}

	public void setRegion(String region) {
		this.region = region;
	}

	public String getWindowCenter() {
		return windowCenter;
	}

	public void setWindowCenter(String windowCenter) {
		this.windowCenter = windowCenter;
	}

	public String getWindowWidth() {
		return windowWidth;
	}

	public void setWindowWidth(String windowWidth) {
		this.windowWidth = windowWidth;
	}

	public String getFrameNumber() {
		return frameNumber;
	}

	public void setFrameNumber(String frameNumber) {
		this.frameNumber = frameNumber;
	}

	public String getImageQuality() {
		return imageQuality;
	}

	public void setImageQuality(String imageQuality) {
		this.imageQuality = imageQuality;
	}

	public String getPresentationUID() {
		return presentationUID;
	}

	public void setPresentationUID(String presentationUID) {
		this.presentationUID = presentationUID;
	}

	public String getPresentationSeriesUID() {
		return presentationSeriesUID;
	}

	public void setPresentationSeriesUID(String presentationSeriesUID) {
		this.presentationSeriesUID = presentationSeriesUID;
	}

	public String getTransferSyntax() {
		return transferSyntax;
	} 

	public void setTransferSyntax(String transferSyntax) {
		this.transferSyntax = transferSyntax;
	}

	public List<String> getContentTypes() {
		return contentTypes;
	}

	public void setContentTypes(String contentType) {
		this.contentTypes = contentType != null ? getDelimetedStringAsList(contentType, ",") : new ArrayList<String>();
	}

	public Attributes getObjectInfo() {
		return objectInfo;
	}

	public void setObjectInfo(Attributes objectInfo) {
		this.objectInfo = objectInfo;
	}
	
    public boolean isExcludePrivate() {
        return "no".equalsIgnoreCase(privateTags);
    }

	public String getPrivateTags() {
		return privateTags;
	}

	public void setPrivateTags(String privateTags) {
		this.privateTags = privateTags;
	}

	public String getSimpleFrameList() {
		return simpleFrameList;
	}

	public void setSimpleFrameList(String simpleFrameList) {
		this.simpleFrameList = simpleFrameList;
	}

	public String getCalculatedFrameList() {
		return calculatedFrameList;
	}

	public void setCalculatedFrameList(String calculatedFrameList) {
		this.calculatedFrameList = calculatedFrameList;
	}

	public void setContentTypes(List<String> contentTypes) {
		this.contentTypes = contentTypes;
	}
	
    public String getRedirectParam() {
		return redirectParam;
	}

	public void setRedirectParam(String redirectParam) {
		this.redirectParam = redirectParam;
	}

	public String getAuthorization() {
		return authorization;
	}

	public void setAuthorization(String authorization) {
		this.authorization = authorization;
	}

	public String getUseOrig() {
		return useOrig;
	}

	public void setUseOrig(String useOrig) {
		this.useOrig = useOrig != null ? useOrig : "true";
	}

	public String getRequestUrl() {
		return requestUrl;
	}

	public void setRequestUrl(String requestUrl) {
		this.requestUrl = requestUrl;
	}
	
	public static List<String> getDelimetedStringAsList(String delimetedString, String delimeter) {
		List<String> returnList = null;
		if (delimetedString == null) {
			return returnList;
		}
		if (delimetedString.length() == 0) {
			return returnList;
		}
		returnList = new ArrayList<String>();
		StringTokenizer tokenizer = new StringTokenizer(delimetedString,
				delimeter);
		while (tokenizer.hasMoreTokens()) {
			returnList.add(tokenizer.nextToken());
		}
		return returnList;
	}
}
