package com.teletip.wado.util;

import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.geom.AffineTransform;
import java.awt.image.AffineTransformOp;
import java.awt.image.BandedSampleModel;
import java.awt.image.BufferedImage;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferUShort;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;

import javax.imageio.ImageIO;
import javax.imageio.ImageReadParam;
import javax.imageio.ImageReader;
import javax.imageio.stream.FileImageInputStream;
import javax.imageio.stream.ImageInputStream;

import org.dcm4che.data.Dataset;
import org.dcm4che.dict.Tags;
import org.dcm4che.dict.UIDs;
import org.dcm4che.imageio.plugins.DcmMetadata;
import org.dcm4cheri.imageio.plugins.DcmImageReadParamImpl;
import org.dcm4cheri.imageio.plugins.SimpleYBRColorSpace;
import org.dcm4cheri.util.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.sun.imageio.plugins.jpeg.JPEGImageWriter;

/**
 * DcmUtil DICOM dosyalarini islemeye yonelik metotlari iceren arac sinifidir.
 * @author	Ayhan Ozan Yilmaz
 * @since	01.12.2016
 */
public class DcmUtil {
    static Logger logger = LoggerFactory.getLogger(DcmUtil.class);
    
    private boolean renderOverlays = false;
    
    private static String imageWriterClass = JPEGImageWriter.class.getName();
    
    private static final int[] EMPTY_INTS = {};
    
    private int[] numberOfStudyBags = EMPTY_INTS;
    
    public static final String CONTENT_TYPE_JPEG = "image/jpeg";
    public static final String CONTENT_TYPE_PNG = "image/png";
    public static final String CONTENT_TYPE_PNG16 = "image/png16";

    public static final String CONTENT_TYPE_DICOM = "application/dicom";
    public static final String CONTENT_TYPE_DICOM_XML = "application/dicom+xml";

    public static final String CONTENT_TYPE_HTML = "text/html";

    private static final String CONTENT_TYPE_XHTML = "application/xhtml+xml";

    public static final String CONTENT_TYPE_XML = "text/xml";

    public static final String CONTENT_TYPE_PLAIN = "text/plain";

    public static final String CONTENT_TYPE_MPEG = "video/mpeg";

    public static final List<String> CONTENT_TYPES = Arrays.asList(
            CONTENT_TYPE_JPEG, CONTENT_TYPE_PNG, CONTENT_TYPE_PNG16,
            CONTENT_TYPE_DICOM, CONTENT_TYPE_DICOM_XML,
            CONTENT_TYPE_HTML, CONTENT_TYPE_XHTML, CONTENT_TYPE_XML, CONTENT_TYPE_PLAIN,
            CONTENT_TYPE_MPEG);
    
    private static final String[] COMPRESSED_TRANSFER_SYNTAXES = {
        UIDs.JPEGBaseline, UIDs.JPEGExtended, UIDs.JPEGLossless14, UIDs.JPEGLossless,
        UIDs.JPEGLSLossless, UIDs.JPEGLSLossy, UIDs.JPEG2000Lossless, UIDs.JPEG2000Lossy,
        UIDs.RLELossless
    };

    public boolean isRenderOverlays() {
        return renderOverlays;
    }

    public void setRenderOverlays(boolean renderOverlays) {
        this.renderOverlays = renderOverlays;
    }
    
    public final String getImageWriterClass() {
        return imageWriterClass;
    }

//    public void setImageWriterClass(String imageWriterClass) {
//        getImageWriterWriter(imageWriterClass).dispose();
//        this.imageWriterClass = imageWriterClass;
//    }

    public static void main(String[] args){
    	//TRIAL #1
//    	try {
//    		File dicomFile = new File("C:\\a.dcm");
//    		BufferedImage image = getImage(dicomFile, 0, "30", "30", null,
//                null, null);
//        	imageQuality = maskNull(imageQuality, imageQuality);
//            //File file = _getImageFile(rows + "-" + columns + "-" + region
//            //        + "-" + windowWidth + "-" + windowCenter + "-" + imageQuality,
//            //        studyUID, seriesUID, instanceUID, suffix, null);
//        	File imageFile = new File("C:\\Apps\\a.jpg");
//			_writeImageFile(image, imageFile, imageQuality);
//		} catch (IOException e) {
//			// TODO Auto-generated catch block
//			e.printStackTrace();
//		}
    	
    	//TRIAL #2
//    	File myDicomFile = new File("c:/a.dcm");
//    	BufferedImage myJpegImage = null;
//    	Iterator<ImageReader> iter = ImageIO.getImageReadersByFormatName("DICOM");
//    	ImageReader reader = (ImageReader) iter.next();
//    	DicomImageReadParam param = (DicomImageReadParam) reader.getDefaultReadParam();
//    	try {
//    		   ImageInputStream iis = ImageIO.createImageInputStream(myDicomFile);
//    		   reader.setInput(iis, false);   
//    		   myJpegImage = reader.read(0, param);   
//    		   iis.close();
//    		   if (myJpegImage == null) {
//    			   System.out.println("\nError: couldn't read dicom image!");
//    			   return;
//    			}
//    		   File myJpegFile = new File("c:/Apps/jpegImage.jpg");   
//    		   OutputStream output = new BufferedOutputStream(new FileOutputStream(myJpegFile));
//    		   JPEGImageEncoder encoder = JPEGCodec.createJPEGEncoder(output);
//    		   encoder.encode(myJpegImage);
//    		   output.close();
//    	} 
//    	catch(IOException e) {
//    	   System.out.println("\nError: couldn't read dicom image!"+ e.getMessage());
//    	   return;
//    	}
    	
    	//TRIAL #3
    	//this is neccessary to initialize image readers
//    	BufferedOutputStream out = null;
//    	InputStream dis = null;
//		try {
//			File myDicomFile = new File("E:/Apps/a.dcm");
//			File myJpegFile = new File("E:/Apps/a_window.jpg");  
//			out = new BufferedOutputStream(new FileOutputStream(myJpegFile));
//			dis = new FileInputStream(myDicomFile);
//		} catch (IOException e) {
//			// TODO Auto-generated catch block
//			e.printStackTrace();
//		}
//    	dicom2JpgConverter(dis, out, "600", "300", 0, 0, null);
    	
    }
    /*
    public static void dicom2JpgConverter(InputStream is, OutputStream out, 
    		String rows, String columns,
            float windowCenter, float windowWidth, 
            String quality){
    	//this is neccessary to initialize image readers
    	//ImageIO.scanForPlugins(); --> this is carried to applistener as initialization is required once
    									//and multiple call causes ConcurrentModificationException
    	BufferedImage myJpegImage = null;
    	//Iterator<ImageReader> iter = ImageIO.getImageReadersByFormatName("DICOM");
    	//ImageReader reader = (ImageReader) iter.next();
    	ImageIO.scanForPlugins();
    	ImageReader reader = getDicomImageReader();
    	DicomImageReadParam param = (DicomImageReadParam) reader.getDefaultReadParam();
    	if (windowWidth != 0 && windowCenter != 0) {
        	param.setAutoWindowing(false);
            param.setWindowCenter(windowCenter);
            param.setWindowWidth(windowWidth);
        } else {
        	param.setAutoWindowing(true);
        }
    	
		try {
			ImageInputStream iis = ImageIO.createImageInputStream(is);
			reader.setInput(iis, false);
			myJpegImage = reader.read(0, param);
			iis.close();
			if (myJpegImage == null) {
				System.out.println("\nError: couldn't read dicom image!");
				return;
			}
			if(!DataUtil.isNullOrEmpty(rows) && !DataUtil.isNullOrEmpty(columns)){
				myJpegImage = resize(myJpegImage, rows, columns, reader.getAspectRatio(0));
			}
			JPEGImageEncoder encoder = JPEGCodec.createJPEGEncoder(out);

			JPEGEncodeParam jep = JPEGCodec.getDefaultJPEGEncodeParam(myJpegImage);
			if(!DataUtil.isNullOrEmpty(quality)){
				jep.setQuality(Float.parseFloat(quality), true);				
			}
			encoder.setJPEGEncodeParam(jep);
			encoder.encode(myJpegImage);
			out.close();
		} catch (IOException e) {
			System.out.println("\nError: couldn't read dicom image!"
					+ e.getMessage());
			return;
		}
    }*/
    
    private static BufferedImage resize(BufferedImage bi, String rows, String columns,
            float aspectRatio) {
        int h = 0;
        int w = 0;
        if (rows == null && columns == null) {
            h = bi.getHeight();
            w = bi.getWidth();
        } else {
            if (rows != null) {
                h = Integer.parseInt(rows);
            }
            if (columns != null) {
                w = Integer.parseInt(columns);
            }
        }
        if (h == 0 || w != 0 && h * aspectRatio > w) {
            h = (int) (w / aspectRatio + .5f);
        } else {
            w = (int) (h * aspectRatio + .5f);
        }
        if (w != bi.getWidth() || h != bi.getHeight()) {
            if (bi.getSampleModel() instanceof BandedSampleModel) {
                // convert RGB color-by-plane to TYPE_INT_RGB, otherwise
                // scaleOp.filter(bi, null) will throw
                // ImagingOpException("Unable to transform src image")
                BufferedImage tmp = new BufferedImage(bi.getWidth(),
                        bi.getHeight(), BufferedImage.TYPE_INT_RGB);
                Graphics2D g = tmp.createGraphics();
                try {
                    g.drawImage(bi, 0, 0, null);
                } finally {
                    g.dispose();
                }
                bi = tmp;
            }
            AffineTransform scale = AffineTransform.getScaleInstance(
                    (double) w / bi.getWidth(), (double) h / bi.getHeight());
            AffineTransformOp scaleOp = new AffineTransformOp(scale,
                    AffineTransformOp.TYPE_NEAREST_NEIGHBOR);
            BufferedImage biDest = scaleOp.filter(bi, null);
            return biDest;
        } else {
            return bi;
        }
    }
    /*
    private static BufferedImage getImage(File file, int frame, String rows,
            String columns, String region, String windowWidth,
            String windowCenter) throws IOException {
        Iterator it = ImageIO.getImageReadersByFormatName("DICOM");
        if (!it.hasNext()) {
            return null; // TODO more useful stuff
        }
        ImageReader reader = (ImageReader) it.next();
        ImageInputStream in = new FileImageInputStream(file);
        try {
            reader.setInput(in, false);
            BufferedImage bi = null;
            Rectangle regionRectangle = null;
            try {
                ImageReadParam param = reader.getDefaultReadParam();
                if (region != null) {
                    String[] ss = StringUtils.split(region, ',');
                    int totWidth = reader.getWidth(0);
                    int totHeight = reader.getHeight(0);

                    int topX = (int) Math.round(Double.parseDouble(ss[0]) * totWidth); // top left X value
                    int topY = (int) Math.round(Double.parseDouble(ss[1]) * totHeight); // top left Y value
                    int botX = (int) Math.round(Double.parseDouble(ss[2]) * totWidth); // bottom right X value
                    int botY = (int) Math.round(Double.parseDouble(ss[3]) * totHeight); // bottom right Y value

                    int w = botX - topX;
                    int h = botY - topY;

                    regionRectangle = new Rectangle(topX, topY, w, h);
                    param.setSourceRegion(regionRectangle);
                }

                Dataset data = ((DcmMetadata) reader.getStreamMetadata()).getDataset();

                if (windowWidth != null && windowCenter != null) {
                    data.putDS(Tags.WindowWidth, windowWidth);
                    data.putDS(Tags.WindowCenter, windowCenter);
                }

                DcmImageReadParamImpl dcmParam = (DcmImageReadParamImpl) param;
                if ((data.get(Tags.WindowCenter) == null)
                    && (data.get(Tags.WindowCenter) == null)
                    && (data.get(Tags.SharedFunctionalGroupsSeq) == null)
                    && (data.get(Tags.RescaleIntercept) == null)
                    && (data.get(Tags.RescaleSlope) == null)) {
                    // these result in full white image
                    // param.setMaskPixelData(false);
                	logger.warn("Possible full white image, investigate!");
                    dcmParam.setAutoWindowing(false);
                }

                bi = reader.read(frame, dcmParam);
            } catch (Exception x) {
            	logger.error("Can't read image:", x);
                return null;
            }

//            if (renderOverlays) {
//                mergeOverlays(bi, reader, frame);
//            }

            if (bi.getColorModel().getColorSpace()
                    instanceof SimpleYBRColorSpace) {
                // convert YBR to RGB to workaround jai-imageio-core issue #173:
                // CLibJPEGImageWriter ignores CororSpace != sRGB
                ColorConvertOp colorConvetOp = new ColorConvertOp(
                        ColorSpace.getInstance(ColorSpace.CS_sRGB), null);
                bi = colorConvetOp.filter(bi, null);
            }
            return resize(bi, rows, columns, reader.getAspectRatio(frame));
        } finally {
            // !!!! without this, we get "too many open files" when generating
            // icons in a tight loop
            in.close();
        }
    }

    private static void createJPEG(BufferedImage bi, File file, float quality)
            throws IOException {
        ImageWriter writer = getImageWriterWriter(imageWriterClass);
        ImageOutputStream out = ImageIO.createImageOutputStream(file);
        try {
            writer.setOutput(out);
            ImageWriteParam iwparam = writer.getDefaultWriteParam();
            iwparam.setCompressionMode(ImageWriteParam.MODE_EXPLICIT);
            String[] compressionTypes = iwparam.getCompressionTypes();
            if (compressionTypes != null && compressionTypes.length > 0) {
                for (int i = 0; i < compressionTypes.length; i++) {
                    if (compressionTypes[i].compareToIgnoreCase("jpeg") == 0) {
                        iwparam.setCompressionType(compressionTypes[i]);
                        break;
                    }
                }
            }
            iwparam.setCompressionQuality(quality);
            writer.write(null, new IIOImage(bi, null, null), iwparam);
        } finally {
            out.close();
            writer.dispose();
        }
    }
    
    private static ImageWriter getImageWriterWriter(String imageWriterClass) {
        for (Iterator writers = ImageIO.getImageWritersByFormatName("JPEG"); writers
                .hasNext();) {
            ImageWriter writer = (ImageWriter) writers.next();
            if (writer.getClass().getName().equals(imageWriterClass)) {
                return writer;
            }
        }
        return null;
        //throw new ConfigurationException("No such ImageWriter - "
        //        + imageWriterClass);
    }
    */

        /**
     * Merge the overlays into the buffered image.
     *
     * The overlay implementation is minimal.
     *
     *
     * Currently
     *
     * <ul>
     * <li>In-Pixel-Data overlays are not supported (they are retired
     * as of the current dicom standard)</li>
     * <li>More than 1 overlay data frame is not supported -
     * there is nothing in the standard, no idea how this is supposed
     * to look.</li>
     * <li>Only the first overlay group is supported. Again,
     * nothing in standard, will implement if given example images.</li>
     * </ul>
     *
     * @param bi
     * @param ds
     */
    /*
    private void mergeOverlays(BufferedImage bi, ImageReader reader, int frame)
            throws IOException {

        Dataset ds = ((DcmMetadata) reader.getStreamMetadata()).getDataset();

        long t1 = System.currentTimeMillis();

        ArrayList<Integer> oldStyleOverlayPlanes = new ArrayList<Integer>();

        for (int group = 0; group < 0x20; group += 2) {

            int oBitPosition = getInt(ds, group, Tags.OverlayBitPosition, -1);
            int oRows = getInt(ds, group, Tags.OverlayRows, -1);
            int oCols = getInt(ds, group, Tags.OverlayColumns, -1);
            int[] oOrigin = getInts(ds, group, Tags.OverlayOrigin);
            int oBitsAllocated = getInt(ds, group, Tags.OverlayBitsAllocated, -1);
            String oType = getString(ds, group, Tags.OverlayType);
            int oNumberOfFrames = getInt(ds, group, 0x60000015, 1);
            int oFrameStart = getInt(ds, group, 0x60000051, 1) - 1;
            int oFrameEnd = oFrameStart + oNumberOfFrames;

            if (oBitPosition == -1 &&
                    oBitsAllocated == -1 &&
                    oRows == -1 &&
                    oCols == -1) {
                logger.trace("No overlay data associated with image for group {}", group);
                continue;
            }


            if ("R".equals(oType)) {
                logger.debug("Overlay ROI bitmap, not doing anything");
                continue;
            }

            if ((oBitsAllocated != 1) && (oBitPosition != 0)) {
                logger.debug("Overlay: {}  OldStyle bitPostion {}", group, oBitPosition);
                oldStyleOverlayPlanes.add(oBitPosition);
                continue;
            }

            if ("GR".indexOf(oType) < 0) {
                logger.warn("mergeOverlays(): Overlay Type {} not supported", oType);
                continue;
            }

            int oX1 = 0;
            int oY1 = 0;
            if (oOrigin != null) {
                oX1 = oOrigin[1] - 1;
                oY1 = oOrigin[0] - 1;
            }

            logger.debug("Overlay: {} OverlayType: {}", group, oType);
            logger.debug("Overlay: {} OverlayRows: {}", group, oRows);
            logger.debug("Overlay: {} OverlayColumns: {}", group, oCols);
            logger.debug("Overlay: {} OverlayOrigin: {} {}", new Object[]{group, oX1, oY1});
            logger.debug("Overlay: {} for Frames: [{}, {})", new Object[]{group, oFrameStart, oFrameEnd});

            if (!((oFrameStart <= frame) && (frame < oFrameEnd))) {
                logger.debug("Overlay: frame {} not in range, skipping", frame);
            }

            int oFrameOffset = frame - oFrameStart;
            int bitOffset = oFrameOffset * oRows * oCols;
            int byteOffset = bitOffset / 8;  // dont round up!
            int numBits = oRows * oCols;
            int numBytes = (numBits + 7) / 8; // round up!

            logger.debug("Overlay: {} bitOffset: {}", group, bitOffset);
            logger.debug("Overlay: {} byteOffset: {}", group, byteOffset);
            logger.debug("Overlay: {} numBits: {}", group, numBits);
            logger.debug("Overlay: {} numBytes: {}", group, numBytes);

            ByteBuffer bb = ds.get(groupedTag(group, Tags.OverlayData)).getByteBuffer();
            logger.debug("Overlay: {} ByteBuffer: {}", group, bb);
            logger.debug("Overlay: {} ByteBuffer ByteOrder: {}", group, (bb.order() == ByteOrder.BIG_ENDIAN) ? "BigEndian" : "LittleEndian");


            IndexColorModel icm = new IndexColorModel(1, 2, icmColorValues, icmColorValues, icmColorValues, 0);
            BufferedImage overBi = new BufferedImage(oCols, oRows, BufferedImage.TYPE_BYTE_BINARY, icm);
            DataBufferByte dataBufferByte = (DataBufferByte) overBi.getRaster().getDataBuffer();


            byte[] dest = dataBufferByte.getData();

            // java awt cant even handle non byte packed lines
            // so if a line lenght is not a multiple of 8, it gets really slow
            int packedRowBits = (oCols + 7) & (~7);
            logger.debug("packed row bits: {}", packedRowBits);

            if(packedRowBits == oCols) {
                // overlay is 8-bit padded, we can do it fast
                for (int i = 0; i < numBytes; i++) {
                    int idx = bb.get(byteOffset + i) & 0xFF;
                    dest[i] = bitSwapLut[idx];
                }
            } else {
                // no joy, slow version
                int packedRowBytes = packedRowBits / 8;
                for (int y = 0; y < oRows; y++) {
                    int rowBitOffset = bitOffset + y * oCols;
                    int rowByteOffset = rowBitOffset / 8;
                    int packedRowByteOffset = y * packedRowBytes;
                    int bitsToMove = (rowBitOffset % 8);
                    if(bitsToMove != 0) {
                        for (int i = 0, size = packedRowBytes - 1; i < size; i++) {
                            int inOffset = rowByteOffset + i;
                            int b1 = bb.get(inOffset) & 0xFF;
                            int b2 = bb.get(inOffset + 1) & 0xFF;
                            int rc = ((b1 >> bitsToMove) ^
                                    ((b2 << (8 - bitsToMove)) & 0xFF));
                            dest[packedRowByteOffset + i] = bitSwapLut[rc];
                        }
                    } else {
                        for(int i = 0, size=packedRowBytes - 1; i < size; i++) {
                            int inOffset = rowByteOffset + i;
                            int b1 = bb.get(inOffset) & 0xFF;
                            dest[packedRowByteOffset + i] = bitSwapLut[b1];
                        }
                    }
                }
            }


            logger.debug("Overlay: {} BufferedImage: {}", group, overBi);
            Graphics2D gBi = bi.createGraphics();
            gBi.drawImage(overBi, oX1, oY1, null);

        }
        logger.debug("Overlay: done combining overlays");


        if (oldStyleOverlayPlanes.size() > 0) {
            try {

                int bitsStored = ds.getInt(Tags.BitsStored, -1);
                short overlayValue = (short) ((1 << bitsStored) - 1);

                DataBuffer _buffer = bi.getRaster().getDataBuffer();
                if (_buffer.getDataType() == DataBuffer.TYPE_SHORT ||
                        _buffer.getDataType() == DataBuffer.TYPE_USHORT) {
                    int mask = 0;
                    for (int i = 0, size = oldStyleOverlayPlanes.size(); i < size; i++) {
                        int bit = oldStyleOverlayPlanes.get(i);
                        mask |= (1 << bit);
                    }

                    // get the image again, this time without windowing/maskpixeldata
                    DcmImageReadParamImpl param = (DcmImageReadParamImpl) reader.getDefaultReadParam();
                    param.setMaskPixelData(false);
                    param.setAutoWindowing(false);
                    BufferedImage oBi = reader.read(frame, param);

                    DataBufferUShort oDataBuffer = (DataBufferUShort) oBi.getRaster().getDataBuffer();
                    short[] src = oDataBuffer.getData();
                    DataBufferUShort dataBuffer = (DataBufferUShort) bi.getRaster().getDataBuffer();
                    short[] dest = dataBuffer.getData();

                    logger.debug("mergeOverlays(): setting oldStyleOverlays, mask used 0x" + Integer.toHexString(mask));
                    for (int i = 0, size = dest.length; i < size; i++) {
                        if ((src[i] & mask) != 0) {
                            dest[i] = overlayValue;
                        }
                    }
                } else {
                    logger.warn("mergeOverlays(): data buffer type {} not supported", _buffer.getDataType());
                }
            } catch (Exception e) {
                logger.error("mergeOverlays(): ERROR", e);
            }
        }

        long t2 = System.currentTimeMillis();

        logger.info("mergeOverlays(): {}ms", t2 -t1);
    }
    
    private static final int getInt(Dataset ds, int group, int tag, int def) {
        return ds.getInt(groupedTag(group, tag), def);
    }
    
     
    public static void sendDicom2JpgFile(InputStream is, OutputStream out)
            throws IOException {
    	//this is neccessary to initialize image readers
    	//ImageIO.scanForPlugins(); --> this is carried to applistener as initialization is required once
    									//and multiple call causes ConcurrentModificationException
    	
    	//dicomFile = new File("c:/a.dcm");
    	BufferedImage myJpegImage = null;
    	Iterator<ImageReader> iter = ImageIO.getImageReadersByFormatName("DICOM");
    	ImageReader reader = (ImageReader) iter.next();
    	DicomImageReadParam param = (DicomImageReadParam) reader.getDefaultReadParam();
    	try {
    		   ImageInputStream iis = ImageIO.createImageInputStream(is);
    		   reader.setInput(iis, false);   
    		   myJpegImage = reader.read(0, param);   
    		   iis.close();
    		   if (myJpegImage == null) {
    			   System.out.println("\nError: couldn't read dicom image!");
    			   return;
    			}
    		   JPEGImageEncoder encoder = JPEGCodec.createJPEGEncoder(out);
    		   encoder.encode(myJpegImage);
    		   out.close();
    	} 
    	catch(IOException e) {
    	   System.out.println("\nError: couldn't read dicom image!"+ e.getMessage());
    	   return;
    	}
    }*/

    /**
     * @param contentType
     * @return
     */
    private String getFileExtension(String contentType) {
        int pos = contentType.indexOf("/");
        String ext = "";
        if (pos != -1) {
            ext = contentType.substring(pos + 1);
            if (ext.equalsIgnoreCase("jpeg"))
                ext = "jpg";
            else if (ext.equalsIgnoreCase("svg+xml"))
                ext = "svg";
            // do some other mapping here;
            ext = "." + ext;
        }
        return ext.toLowerCase();
    }

    /**
     * Writes an image to the given file.
     * 
     * @param image
     *            The image.
     * @param file
     *            The file within this cache to store the image.
     * 
     * @throws IOException
     */
    /*
    private static void _writeImageFile(BufferedImage bi, File file,
            String imageQuality) throws IOException {
        if (!file.getParentFile().exists()) {
            if (!file.getParentFile().mkdirs()) {
                throw new IOException("Can not create directory:"
                        + file.getParentFile());
            }
        }
        try {
            logger.debug("Create JPEG (" + imageQuality
                    + " quality) for WADO request. file: " + file);
            createJPEG(bi, file, Float.parseFloat(imageQuality) / 100);
        } catch (Throwable x) {
        	logger.error("Can not create JPEG for WADO request. file:" + file);
            if (file.exists()) {
                file.delete();
                logger.error("Cache File removed:" + file);
            }
            if (x instanceof IOException) {
                throw (IOException) x;
            }
            IOException ioe = new IOException("Failed to write image file ("
                    + file + ")! Reason:" + x.getMessage());
            ioe.initCause(x);
            throw ioe;
        }
    }
    */
    private static String maskNull(String val, String defval) {
        return val != null ? val : defval;
    }
    
    public static BufferedImage getImage(File file, int frame, String rows,
            String columns, String region, String windowWidth,
            String windowCenter, String contentType, 
            boolean renderOverlays, boolean jpgWriterSupportsByteColormap, boolean jpgWriterSupportsShortColormap) 
            		throws IOException {
        ImageReader reader = getDicomImageReader();
        if (reader == null) {
            return null; // TODO more useful stuff
        }
        ImageInputStream in = new FileImageInputStream(file);
        boolean semaphoreAquired = false;
        try {
            reader.setInput(in, false);
            BufferedImage bi = null;
            Rectangle regionRectangle = null;
            try {
                ImageReadParam param = reader.getDefaultReadParam();
                int totWidth = reader.getWidth(0);
                int totHeight = reader.getHeight(0);
                if (region != null) {
                    String[] ss = StringUtils.split(region, ',');

                    int topX = (int) Math.round(Double.parseDouble(ss[0]) * totWidth); // top left X value
                    int topY = (int) Math.round(Double.parseDouble(ss[1]) * totHeight); // top left Y value
                    int botX = (int) Math.round(Double.parseDouble(ss[2]) * totWidth); // bottom right X value
                    int botY = (int) Math.round(Double.parseDouble(ss[3]) * totHeight); // bottom right Y value

                    int w = botX - topX;
                    int h = botY - topY;

                    regionRectangle = new Rectangle(topX, topY, w, h);
                    param.setSourceRegion(regionRectangle);
                }

                Dataset data = ((DcmMetadata) reader.getStreamMetadata()).getDataset();
                if (windowWidth != null && windowCenter != null) {
                    data.putDS(Tags.WindowWidth, windowWidth);
                    data.putDS(Tags.WindowCenter, windowCenter);
                }
                /*
                if (isCompressed(data)) {
                    semaphoreAquired = DecompressCmd.acquireSemaphore();
                    log.info("start decompression of image: " + totHeight + "x" + totWidth +
                            " (current codec tasks: compress&decompress:" + DecompressCmd.getNrOfConcurrentCodec()+
                            " compress:" + DecompressCmd.getNrOfConcurrentDecompress()+")");
                }*/
                bi = reader.read(frame, param);
            } catch (Exception x) {
            	logger.error("Can't read image:", x);
                return null;
            }

            if (renderOverlays) {
                mergeOverlays(bi, reader, frame);
            }
            return resize(bi, rows, columns, reader.getAspectRatio(frame), contentType, 
            		jpgWriterSupportsByteColormap, jpgWriterSupportsShortColormap);
        } finally {
        	/*
            if (semaphoreAquired) {
                DecompressCmd.finished();
                log.info("finished decompression. (remaining codec tasks: compress&decompress:"+DecompressCmd.getNrOfConcurrentCodec()+
                        " decompress:"+DecompressCmd.getNrOfConcurrentDecompress()+")");
                DecompressCmd.releaseSemaphore();
            }*/
            // !!!! without this, we get "too many open files" when generating
            // icons in a tight loop
            in.close();
        }
    }
    
    private static ImageReader getDicomImageReader() {
        Iterator<ImageReader> it = ImageIO.getImageReadersByFormatName("DICOM");
        while (it.hasNext()) {
            ImageReader reader = it.next();
            if (reader.getClass().getName()
                    //.equals("org.dcm4cheri.imageio.plugins.DcmImageReader"))
            		.equals("org.dcm4che2.imageioimpl.plugins.dcm.DicomImageReader"))
                return reader;
        }
        return null;
    }

    /**
     * Resize the given image.
     * 
     * @param bi
     *                The image as BufferedImage.
     * @param rows
     *                Image height in pixel.
     * @param columns
     *                Image width in pixel.
     * @param aspectRatio 
     * @param contentType 
     * 
     * @return
     */
    private static BufferedImage resize(BufferedImage bi, String rows, String columns,
            float aspectRatio, String contentType, 
            boolean jpgWriterSupportsByteColormap, boolean jpgWriterSupportsShortColormap) {
        int h = 0;
        int w = 0;
        if (rows == null && columns == null) {
            h = bi.getHeight();
            w = bi.getWidth();
        } else {
            if (rows != null) {
                h = Integer.parseInt(rows);
            }
            if (columns != null) {
                w = Integer.parseInt(columns);
            }
        }
        if (h == 0 || w != 0 && h * aspectRatio > w) {
            h = (int) (w / aspectRatio + .5f);
        } else {
            w = (int) (h * aspectRatio + .5f);
        }
        boolean needRGB = !jpgWriterSupportsByteColormap ||
                !jpgWriterSupportsShortColormap && bi.getColorModel().getPixelSize() > 8 ||
                bi.getColorModel().getColorSpace() instanceof SimpleYBRColorSpace;
                boolean rescale = w != bi.getWidth() || h != bi.getHeight();
                logger.debug("Image contentType:"+contentType);
                if (CONTENT_TYPE_JPEG.equals(contentType)) {
                    if (!needRGB && !rescale)
                        return bi;
                    if (needRGB || bi.getSampleModel() instanceof BandedSampleModel) {
                    	logger.debug("Convert BufferedImage to TYPE_INT_RGB!");
                        // convert YBR to RGB to workaround jai-imageio-core issue #173:
                        // CLibJPEGImageWriter ignores CororSpace != sRGB
                        // convert RGB color-by-plane to TYPE_INT_RGB, otherwise
                        // scaleOp.filter(bi, null) will throw
                        // ImagingOpException("Unable to transform src image")
                        BufferedImage tmp = new BufferedImage(bi.getWidth(),
                                bi.getHeight(), BufferedImage.TYPE_INT_RGB);
                        Graphics2D g = tmp.createGraphics();
                        try {
                            g.drawImage(bi, 0, 0, null);
                        } finally {
                            g.dispose();
                        }
                        bi = tmp;
                    }
                } else if (CONTENT_TYPE_PNG.equals(contentType)) {
                    BufferedImage tmp = new BufferedImage(bi.getWidth(),
                            bi.getHeight(), BufferedImage.TYPE_INT_RGB);
                    Graphics2D g = tmp.createGraphics();
                    try {
                        g.drawImage(bi, 0, 0, null);
                    } finally {
                        g.dispose();
                    }
                    bi = tmp;
                } else { //PNG16
                    showMinMaxHist(bi, "Original:");
                    if (bi.getData().getNumBands() == 1) {
                        int pixSize = bi.getColorModel().getPixelSize() > 8 ? 16 : 8;
                        logger.debug("Convert to "+pixSize+"bit GRAY BufferedImage for PNG!");
                        BufferedImage tmp = new BufferedImage(bi.getWidth(), bi.getHeight(), 
                                pixSize == 16 ? BufferedImage.TYPE_USHORT_GRAY : BufferedImage.TYPE_BYTE_GRAY);
                        if (pixSize == 8) {
                            Graphics2D g = tmp.createGraphics();
                            try {
                                g.drawImage(bi, 0, 0, null);
                            } finally {
                                g.dispose();
                            }
                        } else {
                            WritableRaster r = tmp.getRaster();
                            Raster data = bi.getData();
                            for (int x = data.getMinX(), xLen = x+data.getWidth(); x < xLen ; x++) {
                                for (int y = data.getMinY(), yLen = y+data.getHeight(); y < yLen ; y++) {
                                    r.setSample(x, y, 0, data.getSample(x, y, 0));
                                }
                            }
                        }
                        bi = tmp;
                        showMinMaxHist(bi, "Converted:");
                    } else {
                    	logger.debug("########### PNG: use original BufferedImage! numBands > 1");
                    }
                }
                if (!rescale)
                    return bi;

                AffineTransform scale = AffineTransform.getScaleInstance(
                        (double) w / bi.getWidth(), (double) h / bi.getHeight());
                AffineTransformOp scaleOp = new AffineTransformOp(scale,
                        AffineTransformOp.TYPE_NEAREST_NEIGHBOR);
                BufferedImage biDest = scaleOp.filter(bi, null);
                return biDest;
    }
    
    private static void mergeOverlays(BufferedImage bi, ImageReader reader, int frame)
            throws IOException {

        Dataset ds = ((DcmMetadata) reader.getStreamMetadata()).getDataset();

        long t1 = System.currentTimeMillis();

        ArrayList<Integer> oldStyleOverlayPlanes = new ArrayList<Integer>();

        int white = ((1 << ds.getInt(Tags.BitsStored, -1)) - 1);
        for (int group = 0; group < 0x20; group += 2) {
            try {
                int gg0000 = group << 16;
                int oBitPosition = ds.getInt(Tags.OverlayBitPosition | gg0000, -1);
                int oRows = ds.getInt(Tags.OverlayRows | gg0000, -1);
                int oCols = ds.getInt(Tags.OverlayColumns | gg0000, -1);
                int oBitsAllocated = ds.getInt(Tags.OverlayBitsAllocated | gg0000, -1);
                String oType = ds.getString(Tags.OverlayType | gg0000);
                int oNumberOfFrames = ds.getInt(Tags.NumberOfFramesInOverlay | gg0000, 1);
                int oFrameStart = ds.getInt(Tags.ImageFrameOrigin | gg0000, 1) - 1;
                int oFrameEnd = oFrameStart + oNumberOfFrames;

                if (oBitPosition == -1 &&
                        oBitsAllocated == -1 &&
                        oRows == -1 &&
                        oCols == -1) {
                	logger.trace("No overlay data associated with image for group {}", group);
                    continue;
                }


                if ("R".equals(oType)) {
                	logger.debug("Overlay ROI bitmap, not doing anything");
                    continue;
                }

                if ((oBitsAllocated != 1) && (oBitPosition != 0)) {
                	logger.debug("Overlay: {}  OldStyle bitPostion {}", group, oBitPosition);
                    oldStyleOverlayPlanes.add(oBitPosition);
                    continue;
                }

                if ("GR".indexOf(oType) < 0) {
                	logger.warn("mergeOverlays(): Overlay Type {} not supported", oType);
                    continue;
                }



                logger.debug("Overlay: {} OverlayType: {}", group, oType);
                logger.debug("Overlay: {} OverlayRows: {}", group, oRows);
                logger.debug("Overlay: {} OverlayColumns: {}", group, oCols);
                logger.debug("Overlay: {} for Frames: [{}, {})", new Object[]{group, oFrameStart, oFrameEnd});

                if (!((oFrameStart <= frame) && (frame < oFrameEnd))) {
                	logger.debug("Overlay: frame {} not in range, skipping", frame);
                    continue;
                }

                applyOverlay(frame, bi.getRaster(), ds, gg0000, white);
            } catch (Exception x) {
            	logger.warn("Render overlay failed! skipped frame:"+frame+" group:"+group+"!  Enable DEBUG log level to get stacktrace");
            	logger.debug("Reason for skipped overlay:",x);
                continue;
            }
        }
        logger.debug("Overlay: done combining overlays");


        if (oldStyleOverlayPlanes.size() > 0) {
            try {

                int bitsStored = ds.getInt(Tags.BitsStored, -1);
                short overlayValue = (short) ((1 << bitsStored) - 1);

                DataBuffer _buffer = bi.getRaster().getDataBuffer();
                if (_buffer.getDataType() == DataBuffer.TYPE_SHORT ||
                        _buffer.getDataType() == DataBuffer.TYPE_USHORT) {
                    int mask = 0;
                    for (int i = 0, size = oldStyleOverlayPlanes.size(); i < size; i++) {
                        int bit = oldStyleOverlayPlanes.get(i);
                        mask |= (1 << bit);
                    }

                    // get the image again, this time without windowing/maskpixeldata
                    DcmImageReadParamImpl param = (DcmImageReadParamImpl) reader.getDefaultReadParam();
                    param.setMaskPixelData(false);
                    param.setAutoWindowing(false);
                    BufferedImage oBi = reader.read(frame, param);

                    DataBufferUShort oDataBuffer = (DataBufferUShort) oBi.getRaster().getDataBuffer();
                    short[] src = oDataBuffer.getData();
                    DataBufferUShort dataBuffer = (DataBufferUShort) bi.getRaster().getDataBuffer();
                    short[] dest = dataBuffer.getData();

                    logger.debug("mergeOverlays(): setting oldStyleOverlays, mask used 0x" + Integer.toHexString(mask));
                    for (int i = 0, size = dest.length; i < size; i++) {
                        if ((src[i] & mask) != 0) {
                            dest[i] = overlayValue;
                        }
                    }
                } else {
                    logger.warn("mergeOverlays(): data buffer type {} not supported", _buffer.getDataType());
                }
            } catch (Exception e) {
                logger.error("mergeOverlays(): ERROR", e);
            }
        }

        long t2 = System.currentTimeMillis();

        logger.info("mergeOverlays(): {}ms", t2 -t1);
    }

    private static void applyOverlay(int frame, WritableRaster raster,
            Dataset attrs, int gg0000, int pixelValue) {

        int imageFrameOrigin = attrs.getInt(Tags.ImageFrameOrigin | gg0000, 1);
        int framesInOverlay = attrs.getInt(Tags.NumberOfFramesInOverlay
                | gg0000, 1);
        int ovlyFrameIndex = frame - imageFrameOrigin + 1;
        if (ovlyFrameIndex < 0 || ovlyFrameIndex >= framesInOverlay)
            return;

        int tagOverlayRows = Tags.OverlayRows | gg0000;
        int tagOverlayColumns = Tags.OverlayColumns | gg0000;
        int tagOverlayData = Tags.OverlayData | gg0000;
        int tagOverlayOrigin = Tags.OverlayOrigin | gg0000;

        int ovlyRows = attrs.getInt(tagOverlayRows, -1);
        int ovlyColumns = attrs.getInt(tagOverlayColumns, -1);
        int[] ovlyOrigin = attrs.getInts(tagOverlayOrigin);
        byte[] ovlyData = null;

        ovlyData = attrs.get(tagOverlayData).getByteBuffer().array();

        if (ovlyData == null)
            throw new IllegalArgumentException("Missing " + gg0000
                    + " Overlay Data");

        if (ovlyRows <= 0)
            throw new IllegalArgumentException(gg0000 + " Overlay Rows ["
                    + ovlyRows + "]");
        if (ovlyColumns <= 0)
            throw new IllegalArgumentException(gg0000 + " Overlay Columns ["
                    + ovlyColumns + "]");
        if (ovlyOrigin == null)
            throw new IllegalArgumentException("Missing " + gg0000
                    + " Overlay Origin");
        if (ovlyOrigin.length != 2)
            throw new IllegalArgumentException(gg0000 + " Overlay Origin "
                    + Arrays.toString(ovlyOrigin));

        int x0 = ovlyOrigin[1] - 1;
        int y0 = ovlyOrigin[0] - 1;

        int ovlyLen = ovlyRows * ovlyColumns;
        int ovlyOff = ovlyLen * ovlyFrameIndex;
        int numBands = raster.getNumBands();
        for (int i = ovlyOff >>> 3, end = (ovlyOff + ovlyLen + 7) >>> 3; i < end; i++) {
            int ovlyBits = ovlyData[i] & 0xff;
            for (int j = 0; (ovlyBits >>> j) != 0; j++) {
                if ((ovlyBits & (1 << j)) == 0)
                    continue;

                int ovlyIndex = ((i << 3) + j) - ovlyOff;
                if (ovlyIndex >= ovlyLen)
                    continue;

                int y = y0 + ovlyIndex / ovlyColumns;
                int x = x0 + ovlyIndex % ovlyColumns;
                try {
                    for (int b = 0 ; b < numBands ; b++)
                        raster.setSample(x, y, b, pixelValue);
                } catch (ArrayIndexOutOfBoundsException ignore) {
                }
            }
        }
    }
    
    private static void showMinMaxHist(BufferedImage bi, String msg) {
        if (logger.isDebugEnabled()) {
            logger.debug("###############################################################################");
            logger.debug("##########"+msg);
            logger.debug("########## bi type:"+bi.getType());
            logger.debug("########## bi colorModel:"+bi.getColorModel());
            logger.debug("########## bi colorSpace:"+bi.getColorModel().getColorSpace());
            logger.debug("########## bi colorSpaceType:"+bi.getColorModel().getColorSpace().getType());
            logger.debug("########## bi sampleModel:"+bi.getSampleModel());
            logger.debug("########## bi bits/pixel:"+bi.getColorModel().getPixelSize());
            logger.debug("########## bi numBands:"+bi.getData().getNumBands());
            if (logger.isTraceEnabled()) { 
                Raster data = bi.getData();
                HashSet<Integer> hist = new HashSet<Integer>(10000);
                int v, vmax = 0, vmin =Integer.MAX_VALUE;
                for (int x = data.getMinX(), xLen = x+data.getWidth(); x < xLen ; x++) {
                    for (int y = data.getMinY(), yLen = y+data.getHeight(); y < yLen ; y++) {
                        v = data.getSample(x, y, 0);
                        hist.add(v);
                        if (v > vmax) vmax = v;
                        if (v < vmin) vmin = v;
                    }
                }
                logger.trace("##########\n########## samples: min:"+vmin+" max:"+vmax+" nrOfDifferentValues:"+hist.size());
            }
            logger.debug("###############################################################################");
        }
    }
}
